"""Interact with the Censys Search Cert API."""
from typing import List, Optional

from .api import CensysSearchAPIv2


class CensysCerts(CensysSearchAPIv2):
    """Interacts with the Certs index.

    Please note that this class represents only the v2 API endpoints. The v1 API endpoints (search, view, and report) are avilable only from CensysCertificates.

    Examples:
        Inits Censys Certs.

        >>> from censys.search import CensysCerts
        >>> c = CensysCerts()

        Search for hosts by sha256fp.

        >>> c.get_hosts_by_cert("fb444eb8e68437bae06232b9f5091bccff62a768ca09e92eb5c9c2cf9d17c426")
        (
            [
                {
                    "ip": "string",
                    "name": "string",
                    "observed_at": "2021-08-02T14:56:38.711Z",
                    "first_observed_at": "2021-08-02T14:56:38.711Z",
                }
            ],
            {
                "next": "nextCursorToken",
            },
        )
    """

    INDEX_NAME = "certificates"
    """Name of Censys Index."""

    def __init__(
        self, api_id: Optional[str] = None, api_secret: Optional[str] = None, **kwargs
    ):
        """Inits CensysCerts.

        See CensysSearchAPIv2 for additional arguments.

        Args:
            api_id (Optional[str], optional): API ID. Defaults to None.
            api_secret (Optional[str], optional): API Secret. Defaults to None.
            **kwargs: Arbitrary keyword arguments.
        """
        super().__init__(api_id=api_id, api_secret=api_secret, **kwargs)
        self.bulk_path = f"/v2/{self.INDEX_NAME}/bulk"

    def view(self, document_id: str, **kwargs) -> dict:
        """Fetches the certificate record for the specified SHA-256 fingerprint.

        Args:
            document_id (str): The SHA-256 fingerprint of the requested certificate.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            dict: Certificate details.
        """
        return self._get(self.view_path + document_id, args=kwargs)["result"]

    def bulk_post(self, fingerprints: List[str]) -> List[dict]:
        """Fetches the certificate records for the specified SHA-256 fingerprints.

        Using the POST method allows for a larger number of fingerprints to be queried at once.

        Args:
            fingerprints (List[str]): List of certificate SHA256 fingerprints.

        Returns:
            dict: Certificate details.
        """
        data = {"fingerprints": fingerprints}
        return self._post(self.bulk_path, data=data)["result"]

    def bulk_get(self, fingerprints: List[str]) -> List[dict]:
        """Fetches the certificate records for the specified SHA-256 fingerprints.

        Using the GET method allows for a smaller number of fingerprints to be queried at once.

        Args:
            fingerprints (List[str]): List of certificate SHA256 fingerprints.

        Returns:
            dict: Certificate details.
        """
        args = {"fingerprints": fingerprints}
        return self._get(self.bulk_path, args=args)["result"]

    def bulk(self, fingerprints: List[str]) -> List[dict]:
        """Fetches the certificate records for the specified SHA-256 fingerprints.

        By default, this function uses the POST method, which allows for a larger number of fingerprints to be queried at once.
        If you wish to use the GET method, please use `CensysCerts.bulk_get` instead.

        Args:
            fingerprints (List[str]): List of certificate SHA256 fingerprints.

        Returns:
            dict: Certificate details.
        """
        return self.bulk_post(fingerprints)

    def bulk_view(self, fingerprints: List[str]) -> List[dict]:  # type: ignore[override]
        """Fetches the certificate records for the specified SHA-256 fingerprints.

        By default, this function uses the POST method, which allows for a larger number of fingerprints to be queried at once.
        If you wish to use the GET method, please use `CensysCerts.bulk_get` instead.

        Args:
            fingerprints (List[str]): List of certificate SHA256 fingerprints.

        Returns:
            dict: Certificate details.
        """
        return self.bulk_post(fingerprints)

    def search_post_raw(
        self,
        query: str,
        per_page: int = 50,
        cursor: Optional[str] = None,
        fields: Optional[List[str]] = None,
        sort: Optional[List[str]] = None,
        **kwargs,
    ) -> dict:
        """Searches the Certs index using the POST method. Returns the raw response.

        Args:
            query (str): The query string to search for.
            per_page (int): The number of results to return per page. Defaults to 50.
            cursor (str, optional): Cursor token from the API response, which fetches the next page of results when added to the endpoint URL.
            fields (List[str], optional): Additional fields to return in the matched certificates outside of the default returned fields.
            sort (List[str], optional): A list of fields to sort on. By default, fields will be sorted in ascending order.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            dict: Search results.
        """
        data = {
            "q": query,
            "per_page": per_page,
        }
        if cursor:
            data["cursor"] = cursor
        if fields:
            data["fields"] = fields
        if sort:
            data["sort"] = sort
        data.update(kwargs)
        return self._post(self.search_path, data=data)

    def search_post(
        self,
        query: str,
        per_page: int = 50,
        cursor: Optional[str] = None,
        fields: Optional[List[str]] = None,
        sort: Optional[List[str]] = None,
        **kwargs,
    ) -> dict:
        """Searches the Certs index using the POST method.

        This method returns the `result` field of the raw response.
        If you wish to access the raw response, please use `CensysCerts.search_post_raw` instead.

        Args:
            query (str): The query string to search for.
            per_page (int): The number of results to return per page. Defaults to 50.
            cursor (str, optional): Cursor token from the API response, which fetches the next page of results when added to the endpoint URL.
            fields (List[str], optional): Additional fields to return in the matched certificates outside of the default returned fields.
            sort (List[str], optional): A list of fields to sort on. By default, fields will be sorted in ascending order.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            dict: Search results.
        """
        return self.search_post_raw(
            query=query,
            per_page=per_page,
            cursor=cursor,
            fields=fields,
            sort=sort,
            **kwargs,
        )["result"]

    def search_get(
        self,
        query: str,
        per_page: int = 50,
        cursor: Optional[str] = None,
        fields: Optional[List[str]] = None,
        sort: Optional[List[str]] = None,
    ) -> dict:
        """Searches the Certs index using the GET method.

        Args:
            query (str): The query string to search for.
            per_page (int): The number of results to return per page. Defaults to 50.
            cursor (str, optional): Cursor token from the API response, which fetches the next page of results when added to the endpoint URL.
            fields (List[str], optional): Additional fields to return in the matched certificates outside of the default returned fields.
            sort (List[str], optional): A list of fields to sort on. By default, fields will be sorted in ascending order.

        Returns:
            dict: Search results.
        """
        args = {
            "q": query,
            "per_page": per_page,
            "cursor": cursor,
            "fields": fields,
            "sort": sort,
        }
        return self._get(self.search_path, args=args)["result"]

    def raw_search(
        self,
        query: str,
        per_page: int = 50,
        cursor: Optional[str] = None,
        fields: Optional[List[str]] = None,
        sort: Optional[List[str]] = None,
        **kwargs,
    ) -> dict:
        """Searches the Certs index.

        Searches the Certs index for all records that match the given query.
        This method does no automatic pagination or post processing.

        Args:
            query (str): The query string to search for.
            per_page (int): The number of results to return per page. Defaults to 50.
            cursor (str, optional): Cursor token from the API response, which fetches the next page of results when added to the endpoint URL.
            fields (List[str], optional): Additional fields to return in the matched certificates outside of the default returned fields.
            sort (List[str], optional): A list of fields to sort on. By default, fields will be sorted in ascending order.
            **kwargs: Additional keyword arguments to pass to the underlying HTTP request.

        Returns:
            dict: Search results.
        """
        return self.search_post_raw(
            query=query,
            per_page=per_page,
            cursor=cursor,
            fields=fields,
            sort=sort,
            **kwargs,
        )

    def search(  # type: ignore[override]
        self,
        query: str,
        per_page: int = 50,
        cursor: Optional[str] = None,
        pages: int = 1,
        fields: Optional[List[str]] = None,
        sort: Optional[List[str]] = None,
        **kwargs,
    ) -> CensysSearchAPIv2.Query:
        """Searches the Certs index.

        By default, this function uses the POST method, which allows for a larger number of fingerprints to be queried at once.
        If you wish to use the GET method, please use `CensysCerts.search_get` instead.

        Args:
            query (str): The query string to search for.
            per_page (int): The number of results to return per page. Defaults to 50.
            cursor (str, optional): Cursor token from the API response, which fetches the next page of results when added to the endpoint URL.
            pages (int): The number of pages to return. Defaults to 1.
            fields (List[str], optional): Additional fields to return in the matched certificates outside of the default returned fields.
            sort (List[str], optional): A list of fields to sort on. By default, fields will be sorted in ascending order.
            **kwargs: Additional keyword arguments to pass to the underlying HTTP request.

        Returns:
            Query: A query object that can be used to iterate over the search results.
        """
        if fields:
            kwargs["fields"] = fields
        if sort:
            kwargs["sort"] = sort
        return super().search(query, per_page, cursor, pages, **kwargs)

    def aggregate(
        self, query: str, field: str, num_buckets: int = 50, **kwargs
    ) -> dict:
        """Aggregates certificate records matching a specified query into buckets based on the given field.

        Args:
            query (str): The query string to search for.
            field (str): The field to aggregate on.
            num_buckets (int): The number of buckets to return. Defaults to 50.
            **kwargs: Additional keyword arguments to pass to the underlying HTTP request.

        Returns:
            dict: Aggregation results.
        """
        args = {"q": query, "field": field, "num_buckets": num_buckets}
        args.update(kwargs)
        return self._get(self.aggregate_path, args=args)["result"]

    def get_hosts_by_cert(self, fingerprint: str, cursor: Optional[str] = None) -> dict:
        """Returns a list of hosts which contain services presenting this certificate, including when the certificate was first observed.

        Args:
            fingerprint (str): The SHA-256 fingerprint of the requested certificate.
            cursor (str): Cursor token from the API response, which fetches the next page of hosts when added to the endpoint URL.

        Returns:
            dict: A list of hosts which contain services presenting this certificate.
        """
        args = {"cursor": cursor}
        return self._get(self.view_path + fingerprint + "/hosts", args)["result"]

    def list_certs_with_tag(self, tag_id: str) -> List[dict]:
        """Returns a list of certs which are tagged with the specified tag.

        Args:
            tag_id (str): The ID of the tag.

        Returns:
            List[dict]: A list of certs which are tagged with the specified tag.
        """
        return self._list_documents_with_tag(tag_id, "certificates", "certs")



"""Interact with the Censys Web Entities Assets API."""
from typing import Iterator, Optional

from .assets import Assets


class WebEntitiesAssets(Assets):
    """Web Entities Assets API class."""

    def __init__(self, *args, **kwargs):
        """Inits WebEntitiesAssets.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        super().__init__("web-entities", *args, **kwargs)

    def get_assets(self, *args, **kwargs):
        """Requests assets data.

        This method is not implemented for web entities.
        Please see the inventory search and aggregation API.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Raises:
            NotImplementedError: This method is not implemented.
        """
        raise NotImplementedError("This method is not implemented.")

    def get_instances(
        self,
        name_and_port: str,
        page_size: Optional[int] = None,
        cursor: Optional[str] = None,
    ) -> Iterator[dict]:
        """List all instances of the web entity.

        Args:
            name_and_port: (str): Web entity to query.
            page_size (int): Optional; Page size for retrieving assets.
            cursor (str): Optional; Cursor to use for pagination.

        Yields:
            dict: The assets result returned.
        """
        args = {"pageSize": page_size, "cursor": cursor}

        while True:
            res = self._get(f"{self.base_path}/{name_and_port}/instances", args=args)
            yield from res.get("instances", [])
            cursor = res.get("cursor")
            if not cursor:
                break
            args["cursor"] = cursor



"""Interact with the Censys Inventory Search API."""
from typing import List, Optional

from .api import CensysAsmAPI


class InventorySearch(CensysAsmAPI):
    """Inventory Search API class."""

    base_path = "/inventory/v1"

    def search(
        self,
        workspaces: List[str],
        query: Optional[str] = None,
        page_size: Optional[int] = None,
        cursor: Optional[str] = None,
        sort: Optional[List[str]] = None,
        fields: Optional[List[str]] = None,
    ) -> dict:
        """Search inventory data.

        Args:
            workspaces (List[str]): List of workspace IDs to search.
            query (str, optional): Query string.
            page_size (int, optional): Number of results to return. Defaults to 50.
            cursor (str, optional): Cursor to start search from.
            sort (List[str], optional): List of fields to sort by.
            fields (List[str], optional): List of fields to return.

        Returns:
            dict: Inventory search results.
        """
        if page_size is None:
            page_size = 50
        args = {
            "workspaces": workspaces,
            "query": query,
            "pageSize": page_size,
            "cursor": cursor,
            "sort": sort,
            "fields": fields,
        }

        return self._get(self.base_path, args=args)

    def aggregate(
        self,
        workspaces: List[str],
        query: Optional[str] = None,
        aggregation: Optional[dict] = None,
    ) -> dict:
        """Aggregate inventory data.

        Args:
            workspaces (List[str]): List of workspace IDs to search.
            query (str, optional): Query string.
            aggregation (dict, optional): Aggregation object.

        Returns:
            dict: Inventory aggregation results.
        """
        body = {
            "workspaces": workspaces,
            "query": query,
            "aggregation": aggregation,
        }

        return self._post(f"{self.base_path}/aggregate", data=body)

    def fields(self, fields: Optional[List[str]] = None) -> dict:
        """List inventory fields.

        If no fields are specified, all fields will be returned.

        Args:
            fields (List[str], optional): List of fields to return.

        Returns:
            dict: Inventory field results.
        """
        args = {"fields": fields}

        return self._get(f"{self.base_path}/fields", args=args)


Inventory = InventorySearch



"""Censys search CLI."""
import argparse
import json
import sys
import webbrowser
from pathlib import Path
from typing import Any, Dict, List
from urllib.parse import urlencode

from censys.cli.utils import V2_INDEXES, err_console, write_file
from censys.common.exceptions import CensysCLIException
from censys.search import SearchClient
from censys.search.v2.api import CensysSearchAPIv2

Results = List[dict]

DATA_DIR = Path(__file__).parent.parent / "data"
HOSTS_AUTOCOMPLETE = DATA_DIR / "hosts_autocomplete.json"
CERTIFICATES_AUTOCOMPLETE = DATA_DIR / "certificates_autocomplete.json"


def fields_completer(
    prefix: str, parsed_args: argparse.Namespace, **kwargs
) -> List[str]:
    """Fields completer.

    Args:
        prefix (str): Prefix to complete.
        parsed_args (Namespace): Argparse Namespace.
        **kwargs: Keyword arguments.

    Returns:
        List[str]: List of fields.
    """
    index_type = parsed_args.index_type
    autocomplete_json = {}
    if index_type == "hosts":
        with HOSTS_AUTOCOMPLETE.open() as autocomplete_file:
            autocomplete_json = json.load(autocomplete_file)
    elif index_type == "certificates":
        with CERTIFICATES_AUTOCOMPLETE.open() as autocomplete_file:
            autocomplete_json = json.load(autocomplete_file)
    else:
        return []

    autocomplete_data = autocomplete_json.get("data", [])
    fields = [field["value"] for field in autocomplete_data]

    if not prefix:
        # Returns first 20 fields if no prefix is provided
        return fields[:20]
    return fields


def cli_search(args: argparse.Namespace):
    """Search subcommand.

    Args:
        args (Namespace): Argparse Namespace.

    Raises:
        CensysCLIException: If invalid options are provided.
    """
    index_type = args.index_type or args.query_type

    if args.open:
        url_query = {"q": args.query, "resource": index_type}
        webbrowser.open(f"https://search.censys.io/search?{urlencode(url_query)}")
        sys.exit(0)

    censys_args = {}

    if args.api_id:
        censys_args["api_id"] = args.api_id

    if args.api_secret:
        censys_args["api_secret"] = args.api_secret

    c = SearchClient(**censys_args)

    search_args = {}
    write_args = {"file_format": args.format, "file_path": args.output}
    results: List[Dict[str, Any]] = []

    index: CensysSearchAPIv2 = getattr(c.v2, index_type)

    search_args.update(
        {
            "pages": args.pages,
            "per_page": args.per_page,
        }
    )
    if index_type == "hosts":
        search_args.update(
            {
                "sort": args.sort_order,
                "virtual_hosts": args.virtual_hosts,
            }
        )
    elif index_type == "certificates":
        search_args.update(
            {
                "sort": args.sort,
                "fields": args.fields,
            }
        )

    if args.output and not args.output.endswith(".json"):
        raise CensysCLIException(
            "JSON is the only valid file format for Search 2.0 responses."
        )
    write_args.update(
        {
            "file_format": "json" if args.output else "screen",
        }
    )

    with err_console.status("Searching"):
        try:
            for page in index.search(args.query, **search_args):
                results.extend(page)
        except Exception:
            err_console.print_exception(max_frames=4)
        except KeyboardInterrupt:  # pragma: no cover
            pass

    try:
        write_file(results, **write_args)
    except ValueError as error:  # pragma: no cover
        err_console.print(f"Error writing log file. Error: {error}")


def include(parent_parser: argparse._SubParsersAction, parents: dict):
    """Include this subcommand into the parent parser.

    Args:
        parent_parser (argparse._SubParsersAction): Parent parser.
        parents (dict): Parent arg parsers.
    """
    search_parser = parent_parser.add_parser(
        "search",
        description="Query Censys Search for resource data by providing a query \
            string, the resource index, and the fields to be returned",
        help="query Censys search",
        parents=[parents["auth"]],
    )
    search_parser.add_argument(
        "query",
        type=str,
        help="a string written in Censys Search syntax",
    ).completer = fields_completer

    index_metavar = "|".join(V2_INDEXES)
    index_default = "hosts"
    search_parser.add_argument(
        "--index-type",
        type=str,
        default=index_default,
        choices=V2_INDEXES,
        metavar=index_metavar,
        help="which resource index to query",
    )
    # Backwards compatibility
    search_parser.add_argument(
        "-f",
        "--format",
        type=str,
        default="screen",
        choices=["screen", "json"],
        metavar="screen|json",
        help=argparse.SUPPRESS,
    )
    search_parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="output file path",
    )
    search_parser.add_argument(
        "-O",
        "--open",
        action="store_true",
        help="open query in browser",
    )

    search_parser.add_argument(
        "--pages",
        default=1,
        type=int,
        help="number of pages of results to return (when set to -1 returns all pages available)",
    )
    search_parser.add_argument(
        "--per-page",
        default=100,
        type=int,
        help="number of results to return per page",
    )

    hosts_group = search_parser.add_argument_group("hosts specific arguments")
    hosts_group.add_argument(
        "--sort-order",
        dest="sort_order",
        type=str,
        default="RELEVANCE",
        choices=["RELEVANCE", "ASCENDING", "DESCENDING", "RANDOM"],
        help="sort order of results",
    )
    hosts_group.add_argument(
        "--virtual-hosts",
        type=str,
        default="EXCLUDE",
        choices=["INCLUDE", "EXCLUDE", "ONLY"],
        metavar="INCLUDE|EXCLUDE|ONLY",
        help="whether to include virtual hosts in the results",
    )

    certs_group = search_parser.add_argument_group("certificates specific arguments")
    certs_group.add_argument(
        "--sort",
        dest="sort",
        type=str,
        nargs="+",
        help="fields to sort by",
    )
    certs_group.add_argument(
        "--fields",
        dest="fields",
        type=str,
        nargs="+",
        help="additional fields to return in the matched certificates",
    ).completer = fields_completer

    search_parser.set_defaults(func=cli_search)



"""Interact with the Censys Seeds, Assets, and Logbook APIs."""
from typing import Optional

from .assets import (
    CertificatesAssets,
    DomainsAssets,
    HostsAssets,
    SubdomainsAssets,
    WebEntitiesAssets,
)
from .clouds import Clouds
from .events import Events
from .inventory import InventorySearch
from .risks import Risks
from .seeds import Seeds


class AsmClient:
    """Client ASM API class."""

    def __init__(self, api_key: Optional[str] = None, **kwargs):
        """Inits AsmClient.

        Args:
            api_key (str): Optional; The API Key provided by Censys.
            **kwargs: Arbitrary keyword arguments.
        """
        self.seeds = Seeds(api_key, **kwargs)
        self.hosts = HostsAssets(api_key, **kwargs)
        self.certificates = CertificatesAssets(api_key, **kwargs)
        self.domains = DomainsAssets(api_key, **kwargs)
        self.events = Events(api_key, **kwargs)
        self.clouds = Clouds(api_key, **kwargs)
        self.risks = Risks(api_key, **kwargs)
        self.inventory = InventorySearch(api_key, **kwargs)
        self.web_entities = WebEntitiesAssets(api_key, **kwargs)

        # Save the arguments for parameterized client usage
        self.__api_kwargs = kwargs

    def get_subdomains(self, domain: str):
        """Get an API instance for subdomains of the parent domain.

        Args:
            domain: (str): Parent domain to access.

        Returns:
            SubdomainsAssets: A Subdomains Assets API instance .
        """
        return SubdomainsAssets(domain, self.domains._api_key, **self.__api_kwargs)
